@@ -4,14 +4,12 @@ import os
 import sys
 import re
 import math
+import warnings
 
 from collections import defaultdict
 
 import pandas as pd
 import numpy as np
-import seaborn as sns
-import matplotlib.pyplot as plt
-
 import straw
 import cooler
 
@@ -20,6 +18,7 @@ from scipy.ndimage import gaussian_filter
 from scipy.ndimage.filters import maximum_filter
 from scipy.signal import convolve2d
 import scipy.ndimage.measurements as scipy_measurements
+from scipy import sparse
 
 from statsmodels.stats.multitest import multipletests
 
@@ -81,6 +80,14 @@ def parse_args(args):
                         help="RECOMMENDED: biases calculated by\
                         ICE or KR norm for each locus for contact map are read from BIASFILE",
                         required=False)
+    parser.add_argument(
+        "-st",
+        "--sparsityThreshold",
+        dest="st",
+        type=float,
+        default=0.88,
+        help="OPTIONAL: Mustache filters out contacts in sparse areas, you can relax this for sparse datasets(i.e. -st 0.8). Default value is 0.88.",
+        required=False)
     parser.add_argument(
         "-pt",
         "--pThreshold",
@@ -111,6 +118,12 @@ def parse_args(args):
                         required=False)
     parser.add_argument("-p", "--processes", dest="nprocesses", default=4, type=int,
                         help="OPTIONAL: Number of parallel processes to run. DEFAULT is 4. Increasing this will also increase the memory usage", required=False)
+    # parser.add_argument("-c",
+                        # "--changefile",
+                        # dest="changedir",
+                        # help="...",
+                        # required=False,
+                        # default="")					
     parser.add_argument(
         "-ch",
         "--chromosome",
@@ -118,6 +131,13 @@ def parse_args(args):
         help="REQUIRED: Specify which chromosome to run the program for.",
         default='n',
         required=True)
+    parser.add_argument(
+        "-ch2",
+        "--chromosome2",
+        dest="chromosome2",
+        help="Optional: Specify the second chromosome for interchromosomal analysis.",
+        default='n',
+        required=False)
     parser.add_argument("-v",
                         "--verbose",
                         dest="verbose",
@@ -146,71 +166,157 @@ def is_chr(s, c):
     return str(c) in re.findall("[1-9][0-9]*", s)
 
 
+def get_sep(f):
+    """
+    :param f: file path
+    :return: Guesses the value separator in the file.
+    """
+    with open(f) as file:
+        for line in file:
+            if "\t" in line:
+                return '\t'
+            if " " in line.strip():
+                return ' '
+            if "," in line:
+                return ','
+            if len(line.split(' '))==1:
+                return ' '
+            break
+    raise FileNotFoundError
+
+
+def read_bias(f, chromosome, res):
+    """
+    :param f: Path to the bias file
+    :return: Dictionary where keys are the bin coordinates and values are the bias values.
+    """
+    d = defaultdict(lambda: 1.0)
+    if f:
+        sep = get_sep(f)
+        with open(f) as file:
+            for pos,line in enumerate(file):
+                line = line.strip().split(sep)
+                if len(line)==3:
+                    if is_chr(line[0], chromosome):
+                        val = float(line[2])
+                        if not np.isnan(val):
+                            if val<0.2:
+                                d[(float(line[1]) // res)] = np.Inf
+                            else:
+                                d[(float(line[1]) // res)] = val
+                        else:
+                            d[(float(line[1]) // res)] = np.Inf
+							
+                elif len(line)==1:
+                    val = float(line[0])
+                    if not np.isnan(val):
+                        if val<0.2:
+                            d[pos] = np.Inf
+                        else:
+                            d[pos] = val
+                    else:
+                       d[pos] = np.Inf
+				
+				
+        return d
+    return False
+
+
 def read_pd(f, distance, res, bias, chromosome):
-    df = pd.read_csv(f, sep='\t', header=None)
+    sep = get_sep(f)
+    df = pd.read_csv(f, sep=sep, header=None)
     df.dropna(inplace=True)
-    df = df[np.vectorize(is_chr)(df[0], chromosome)]
-    df = df[np.vectorize(is_chr)(df[2], chromosome)]
-    df = df.loc[np.abs(df[1]-df[3]) <= ((distance+10) * res), :]
-    df[1] //= res
-    df[3] //= res
+    if df.shape[1]==5:
+        df = df[np.vectorize(is_chr)(df[0], chromosome)]
+        df = df[np.vectorize(is_chr)(df[2], chromosome)]
+        df = df.loc[np.abs(df[1]-df[3]) <= ((distance+10) * res), :]
+        df[1] //= res
+        df[3] //= res
+        bias = read_bias(bias, chromosome, res)
+        if bias:
+            factors = np.vectorize(bias.get)(df[1], 1)
+            df[4] = np.divide(df[4], factors)
+            factors = np.vectorize(bias.get)(df[3], 1)
+            df[4] = np.divide(df[4], factors)
 
-    if bias:
-        bdf = pd.read_csv(bias, sep='\t', header=None)
-        biases = np.nan_to_num(bdf[0], nan=np.Inf)
-        biases[biases < 0.2] = np.Inf
-        df[4] = df[4] / (biases[df[1]] * biases[df[3]])
+        df = df.loc[df[4] > 0, :]
 
-    df = df.loc[df[4] > 0, :]
+        x = np.min(df.loc[:, [1, 3]], axis=1)
+        y = np.max(df.loc[:, [1, 3]], axis=1)
+        val = np.array(df[4])
+    
+    elif df.shape[1]==3:
+        df = df.loc[np.abs(df[1]-df[0]) <= ((distance+10) * res), :]
+        df[0] //= res
+        df[1] //= res
+        bias = read_bias(bias, chromosome, res)
+        if bias:
+            factors = np.vectorize(bias.get)(df[0], 1)
+            df[2] = np.divide(df[2], factors)
+            factors = np.vectorize(bias.get)(df[1], 1)
+            df[2] = np.divide(df[2], factors)
 
-    x = np.min(df.loc[:, [1, 3]], axis=1)
-    y = np.max(df.loc[:, [1, 3]], axis=1)
+        df = df.loc[df[2] > 0, :]
 
-    n = max(y) + 1
-    c = np.zeros((n, n), dtype=np.float32)
-    c[x, y] = df[4]
-    return c, n
+        x = np.min(df.loc[:, [0, 1]], axis=1)
+        y = np.max(df.loc[:, [0, 1]], axis=1)
+        val = np.array(df[2])  
+    
+    return x,y,val
 
 
-def read_hic_file(f, chr, res):
+def read_hic_file(f, chr, chr2, res):
     """
     :param f: .hic file path
     :param chr: Which chromosome to read the file for
     :param res: Resolution to extract information from
     :return: Numpy matrix of contact counts
     """
-    result = straw.straw('KR', f, str(chr), str(chr), 'BP', res)
+    result = straw.straw('KR', f, str(chr), str(chr2), 'BP', res)
     x = np.array(result[0]) // res
     y = np.array(result[1]) // res
     val = np.array(result[2])
-    dist_f = np.abs(x-y) <= (2_000_000 / res)
-    x = x[dist_f]
-    y = y[dist_f]
-    val = val[dist_f]
-    n = max(max(x), max(y)) + 1
-    o = np.zeros((n, n))
-    o[x, y] = val
-    o[y, x] = val
-    o = np.triu(o)
-    np.nan_to_num(o, copy=False, nan=0, posinf=0, neginf=0)
-    o[np.triu_indices_from(o, k=(2000000+1))] = 0
-    return np.triu(o), n
+    val[np.isnan(val)] = 0
+    if(chr==chr2):
+        dist_f = np.logical_and(np.abs(x-y) <= (2000000 / res),val > 0)
+        x = x[dist_f]
+        y = y[dist_f]
+        val = val[dist_f]
 
+    return np.array(x),np.array(y),np.array(val)
 
-def read_cooler(f, chr):
+def read_cooler(f, chr, chr2,res):
     """
     :param f: .cool file path
     :param chr: Which chromosome to read the file for
     :return: Numpy matrix of contact counts
     """
     clr = cooler.Cooler(f)
-    result = clr.matrix(balance=True).fetch(chr)
+    if chr == chr2:
+        result = clr.matrix(balance=True,sparse=True).fetch(chr) #,as_pixels=True, join=True
+    else:
+        result = clr.matrix(balance=True,sparse=True).fetch(chr, chr2)
+   
+    result = sparse.triu(result)
     np.nan_to_num(result, copy=False, nan=0, posinf=0, neginf=0)
-    result[np.triu_indices_from(result, k=(2000000+1))] = 0
-    return np.triu(result), result.shape[1]
+    x = result.row
+    y = result.col
+    val = result.data
+	
+	# result = result[np.logical_not(np.isnan(result['count']))]
+    # x = np.array(result['start1']) // res
+    # y = np.array(result['start2']) // res
+    # val = np.array(result['count'])
 
+    if(chr==chr2):
+        dist_f = np.abs(x-y) <= (2000000 / res)
+        x = x[dist_f]
+        y = y[dist_f]
+        val = val[dist_f]
 
-def read_mcooler(f, chr, res):
+    return np.array(x),np.array(y),np.array(val)
+
+def read_mcooler(f, chr, chr2, res):
     """
     :param f: .cool file path
     :param chr: Which chromosome to read the file for
@@ -219,11 +325,24 @@ def read_mcooler(f, chr, res):
     """
     uri = '%s::/resolutions/%s' % (f, res)
     clr = cooler.Cooler(uri)
-    result = clr.matrix(balance=True).fetch(chr)
+    if chr == chr2:
+        result = clr.matrix(balance=True,sparse=True).fetch(chr)#as_pixels=True, join=True
+    else:
+        result = clr.matrix(balance=True,sparse=True).fetch(chr, chr2)
+    
+    result = sparse.triu(result)
     np.nan_to_num(result, copy=False, nan=0, posinf=0, neginf=0)
-    result[np.triu_indices_from(result, k=(2000000+1))] = 0
-    return np.triu(result), result.shape[1]
-
+    x = result.row
+    y = result.col
+    val = result.data
+	
+    if(chr==chr2):
+        dist_f = np.abs(x-y) <= (2000000 / res)
+        x = x[dist_f]
+        y = y[dist_f]
+        val = val[dist_f]
+	
+    return np.array(x),np.array(y),np.array(val)
 
 def get_diags(map):
     """
@@ -252,58 +371,68 @@ def get_diags(map):
     return means, stds
 
 
-def normalize_sparse(x, y, v, cmap, resolution, distance):
-    distances = np.abs(y-x)
-    filter_size = int(2_000_000 / resolution)
-    for d in range(2 + distance):
-        indices = distances == d
-        i = kth_diag_indices(cmap, d)
-        vals = np.zeros_like(cmap[i])
-        vals[x[indices]] = v[indices]+0.001
-        if vals.size == 0:
-            continue
-        std = np.std(v[indices])
-        mean = np.mean(v[indices])
-        if math.isnan(mean):
-            mean = 0
-        if math.isnan(std):
-            std = 1
+def normalize_sparse(x, y, v, resolution, distance):
+    n = max(max(x),max(y)) + 1
+    with warnings.catch_warnings():
+        warnings.simplefilter('ignore', category=RuntimeWarning)
+        distances = np.abs(y-x)
+        filter_size = int(2000000 / resolution)
+        for d in range(2 + distance):
+            indices = distances == d
+            vals = np.zeros(n-d)
+            vals[x[indices]] = v[indices]+0.001
+            if vals.size == 0:
+                continue
+            std = np.std(v[indices])
+            mean = np.mean(v[indices])
+            if math.isnan(mean):
+                mean = 0
+            if math.isnan(std):
+                std = 1
 
-        kernel = np.ones(filter_size)
-        counts = np.convolve(vals != 0, kernel, mode='same')
+            kernel = np.ones(filter_size)
+            counts = np.convolve(vals != 0, kernel, mode='same')
 
-        s = np.convolve(vals, kernel, mode='same')
-        s2 = np.convolve(vals ** 2, kernel, mode='same')
-        local_var = (s2 - s ** 2 / counts) / (counts - 1)
+            s = np.convolve(vals, kernel, mode='same')
+            s2 = np.convolve(vals ** 2, kernel, mode='same')
+            local_var = (s2 - s ** 2 / counts) / (counts - 1)
 
-        std2 = std ** 2
-        np.nan_to_num(local_var, copy=False,
-                      neginf=std2, posinf=std2, nan=std2)
+            std2 = std ** 2
+            np.nan_to_num(local_var, copy=False,
+                          neginf=std2, posinf=std2, nan=std2)
 
-        local_mean = s / counts
-        local_mean[counts < 30] = mean
-        local_var[counts < 30] = std2
+            local_mean = s / counts
+            local_mean[counts < 30] = mean
+            local_var[counts < 30] = std2
+			
+            np.nan_to_num(local_mean, copy=False,
+                          neginf=mean, posinf=mean, nan=mean)
 
-        np.nan_to_num(local_mean, copy=False,
-                      neginf=mean, posinf=mean, nan=mean)
-
-        local_std = np.sqrt(local_var)
-        vals[x[indices]] -= local_mean[x[indices]]
-        vals[x[indices]] /= local_std[x[indices]]
-        np.nan_to_num(vals, copy=False, nan=0, posinf=0, neginf=0)
-        vals = vals*(1 + math.log(1+mean, 30))
-        cmap[i] = vals
+            local_std = np.sqrt(local_var)
+            vals[x[indices]] -= local_mean[x[indices]]
+            vals[x[indices]] /= local_std[x[indices]]
+            np.nan_to_num(vals, copy=False, nan=0, posinf=0, neginf=0)
+            vals = vals*(1 + math.log(1+mean, 30))
+            
+            v[indices] = vals[x[indices]]
+			
+def inter_normalize_map(vals):
+    m = np.mean(vals)
+    s = np.std(vals)
+    cmap -= m
+    cmap /= s
+    np.nan_to_num(cmap, copy=False, nan=0, posinf=0, neginf=0)
 
 
-def mustache(cc, chromosome, res, start, end, mask_size, distance, octave_values):
-
-    c = np.copy(cc[start:end, start:end])
+def mustache(c, chromosome,chromosome2, res, start, end, mask_size, distance, octave_values, st):
 
     nz = np.logical_and(c != 0, np.triu(c, 4))
     if np.sum(nz) < 50:
         return []
     c[np.tril_indices_from(c, 4)] = 2
-    c[np.triu_indices_from(c, k=(distance+1))] = 2
+
+    if chromosome == chromosome2:
+        c[np.triu_indices_from(c, k=(distance+1))] = 2
 
     pAll = np.ones_like(c[nz]) * 2
     Scales = np.ones_like(pAll)
@@ -391,7 +520,7 @@ def mustache(cc, chromosome, res, start, end, mask_size, distance, octave_values
         s = 2*s
         c2 = np.sum(nz[x[i]-s:x[i]+s+1, y[i]-s:y[i]+s+1]) / \
             ((2*s+1)**2)
-        if c1 < 0.88 or c2 < 0.6:
+        if c1 < st or c2 < 0.6:
             nonsparse[i] = False
     x = x[nonsparse]
     y = y[nonsparse]
@@ -404,13 +533,13 @@ def mustache(cc, chromosome, res, start, end, mask_size, distance, octave_values
 
     def diag_mean(k, map):
         return nz_mean(map[kth_diag_indices(map, k)])
-
-    means = np.vectorize(diag_mean, excluded=['map'])(k=y-x, map=c)
-    passing_indices = c[x, y] > 2*means
-    if len(passing_indices) == 0 or np.sum(passing_indices) == 0:
-        return []
-    x = x[passing_indices]
-    y = y[passing_indices]
+    if chromosome == chromosome2:
+        means = np.vectorize(diag_mean, excluded=['map'])(k=y-x, map=c)
+        passing_indices = c[x, y] > 2*means
+        if len(passing_indices) == 0 or np.sum(passing_indices) == 0:
+            return []
+        x = x[passing_indices]
+        y = y[passing_indices]
 
     label_matrix = np.zeros((np.max(y)+2, np.max(y)+2), dtype=np.float32)
     label_matrix[x, y] = o[x, y] + 1
@@ -439,12 +568,22 @@ def regulator(f, outdir, bed="",
               res=5000,
               sigma0=1.6,
               s=10,
+              st=0.88,
               octaves=2,
               verbose=True,
               nprocesses=4,
               distance_filter=2000000,
               bias=False,
-              chromosome='n'):
+              chromosome='n',
+			  chromosome2=None):
+    
+    if not chromosome2 or chromosome2 == 'n':
+        chromosome2 = chromosome
+
+    if (chromosome != chromosome2) and not ((('.hic' in f) or ('.cool' in f) or ('.mcool' in f))):
+        print(
+            "Interchromosomal analysis is only supported for .hic and .cool input formats.")
+        raise FileNotFoundError
 
     octave_values = [sigma0 * (2 ** i) for i in range(octaves)]
     distance = distance_filter
@@ -454,52 +593,68 @@ def regulator(f, outdir, bed="",
     print("Reading contact map...")
 
     if f.endswith(".hic"):
-        c, n = read_hic_file(f, chromosome, res)
+        x, y, v = read_hic_file(f, chromosome,chromosome2, res)
     elif f.endswith(".cool"):
-        c, n = read_cooler(f, chromosome)
+        x, y, v = read_cooler(f, chromosome,chromosome2, res)
     elif f.endswith(".mcool"):
-        c, n = read_mcooler(f, chromosome, res)
+        x, y, v = read_mcooler(f, chromosome,chromosome2, res)
     else:
-        c, n = read_pd(f, distance, res, bias, chromosome)
+        x, y, v = read_pd(f, distance, res, bias, chromosome)
 
     print("Normalizing contact map...")
-    x, y = np.nonzero(c)
-    normalize_sparse(x, y, c[x, y], c, res, distance)
-    CHUNK_SIZE = max(2*distance, 2000)
-    overlap_size = distance
+    
+    if chromosome == chromosome2:
+        n = max(max(x), max(y)) + 1
+        normalize_sparse(x, y, v, res, distance)
+        CHUNK_SIZE = max(2*distance, 2000)
+        overlap_size = distance
 
-    if n <= CHUNK_SIZE:
-        start = [0]
-        end = [n]
+        if n <= CHUNK_SIZE:
+            start = [0]
+            end = [n]
+        else:
+            start = [0]
+            end = [CHUNK_SIZE]
+
+            while end[-1] < n:
+                start.append(end[-1]-overlap_size)
+                end.append(start[-1] + CHUNK_SIZE)
+            end[-1] = n
+            start[-1] = end[-1] - CHUNK_SIZE
+
+        print("Loop calling...")
+        with Manager() as manager:
+            o = manager.list()
+            i = 0
+            processes = []
+            for i in range(len(start)):
+                # create the currnet block
+                indx = np.logical_and.reduce((x >= start[i], x < end[i], y >= start[i],y < end[i]))
+                xc = x[indx] - start[i]
+                yc = y[indx] - start[i]
+                vc = v[indx]
+                cc = np.zeros((CHUNK_SIZE,CHUNK_SIZE))
+                cc[xc,yc] = vc
+                #             
+                p = Process(target=process_block, args=(
+                    i, start, end, overlap_size, cc, chromosome,chromosome2, res, distance, octave_values, o, st))
+                p.start()
+                processes.append(p)
+                if len(processes) >= nprocesses or i == (len(start) - 1):
+                    for p in processes:
+                        p.join()
+                    processes = []
+            return list(o)
+        
     else:
-        start = [0]
-        end = [CHUNK_SIZE]
+        n1 = max(x) + 1
+        n2 = max(y) + 1
+        inter_normalize_map(v)
+	
+ 
 
-        while end[-1] < n:
-            start.append(end[-1]-overlap_size)
-            end.append(start[-1] + CHUNK_SIZE)
-        end[-1] = n
-        start[-1] = end[-1] - CHUNK_SIZE
-
-    print("Loop calling...")
-    with Manager() as manager:
-        o = manager.list()
-        i = 0
-        processes = []
-        for i in range(len(start)):
-            p = Process(target=process_block, args=(
-                i, start, end, overlap_size, c, chromosome, res, distance, octave_values, o))
-            p.start()
-            processes.append(p)
-            if len(processes) >= nprocesses or i == (len(start) - 1):
-                for p in processes:
-                    p.join()
-                processes = []
-        return list(o)
-
-
-def process_block(i, start, end, overlap_size, c, chromosome, res, distance, octave_values, o):
-    print(f"Starting block {i+1}/{len(start)}...")
+def process_block(i, start, end, overlap_size, cc, chromosome,chromosome2, res, distance, octave_values, o, st):
+    print("Starting block ", i+1, "/", len(start), "...", sep='')
     if i == 0:
         mask_size = -1
     elif i == len(start)-1:
@@ -507,11 +662,11 @@ def process_block(i, start, end, overlap_size, c, chromosome, res, distance, oct
     else:
         mask_size = overlap_size
     loops = mustache(
-        c, chromosome, res, start[i], end[i], mask_size, distance, octave_values)
+        cc, chromosome,chromosome2, res, start[i], end[i], mask_size, distance, octave_values, st)
     for loop in list(loops):
         if loop[0] >= start[i]+mask_size or loop[1] >= start[i]+mask_size:
             o.append([loop[0], loop[1], loop[2], loop[3]])
-    print(f"Block {i+1} done.")
+    print("Block", i+1, "done.")
 
 
 def main():
@@ -530,7 +685,7 @@ def main():
     if not res:
         print("Error: Invalid resolution")
         return
-    distFilter = 2_000_000
+    distFilter = 2000000
 
     biasf = False
     if args.biasfile:
@@ -539,17 +694,20 @@ def main():
         else:
             print("Error: Couldn't find specified bias file")
             return
-
+    if not args.chromosome2 or args.chromosome2 == 'n':
+        args.chromosome2 = args.chromosome
     o = regulator(f, args.outdir,
                   bed=args.bed,
                   res=res,
                   sigma0=args.s_z,
                   s=args.s,
                   verbose=args.verbose,
+                  st=args.st,
                   distance_filter=distFilter,
                   nprocesses=args.nprocesses,
                   bias=biasf,
                   chromosome=args.chromosome,
+				  chromosome2=args.chromosome2,
                   octaves=args.octaves)
     with open(args.outdir, 'w') as out_file:
         out_file.write(
@@ -557,7 +715,7 @@ def main():
         for significant in o:
             if float(significant[2]) < args.pt:
                 out_file.write(
-                    f'{args.chromosome}\t{significant[0]*res}\t{(significant[0]+1)*res}\t{args.chromosome}\t{significant[1]*res}\t{(significant[1]+1)*res}\t{significant[2]}\t{significant[3]}\n')
+                    str(args.chromosome)+'\t' + str(significant[0]*res) + '\t' + str((significant[0]+1)*res) + '\t' + str(args.chromosome2) + '\t' + str(significant[1]*res) + '\t' + str((significant[1]+1)*res) + '\t' + str(significant[2]) + '\t' + str(significant[3]) + '\n')
 
 
 if __name__ == '__main__':
